__copyright__ = "Copyright (c) 2014 CollabNet, Inc. All rights reserved."


import os
import re
import logging
import urllib2
import time

from app.scripts import BaseInitializationScript
import app.util
import service.util
from suds.client import Client
from suds.cache import ObjectCache

class InitializationScript(BaseInitializationScript):

    scriptDescription = "Register SCM integration"

    def __isRegistered(self, repoBase):
        scmPropertiesFile = os.path.join(repoBase, '.scm.properties')
        if os.path.exists(scmPropertiesFile):
            fileContent = open(scmPropertiesFile, 'r').read()
            if re.search('external_system_id', fileContent):
                self.log.info("SCM integration file found at %s" % scmPropertiesFile)
                return True

        return False
        
    def __portSegment(self, scheme, port):
        if scheme == "https":
            if port == 443:
                return ""
            else:
                return ":%d" % port
        else:
            if port == 80:
                return ""
            else:
                return ":%d" % port
                
    def __getService(self, soapBaseUrl, name):
        sudsCacheDir = os.path.join("%s" %self.cfg['TEMPORARY_DIR'], "suds")
        sudsCache = ObjectCache(sudsCacheDir)
        
        sudsLogFile = os.path.join(self.cfg['RUNTIME_LOG_DIR'], 'suds.log')
        logging.basicConfig(filename=sudsLogFile)
        logging.getLogger('suds.client').setLevel(self.log.getEffectiveLevel())
        
        wsdl = "%s/%s?wsdl" % (soapBaseUrl, name)
        return Client(wsdl, cache=sudsCache, location=wsdl)
        
    def __registerScmIntegration(self, soapBaseUrl, adapterType, name, title, paramNames, paramValues):
        try:
            sessionId = self.__getService(soapBaseUrl, "CollabNet").service.login("scmadmin", self.cfg["SCM_ADMIN_PASSWORD"])
        except Exception, ex:
            raise Exception("Authenticating with TeamForge (at %s) failed: %r" % (soapBaseUrl, ex))
                
        try:
            scmService = self.__getService(soapBaseUrl, "ScmApp")
            addExtSystemSoapDO = scmService.factory.create('ns1:SoapNamedValues')
            addExtSystemSoapDO.names = paramNames
            addExtSystemSoapDO.values = paramValues
            
            exsysid = scmService.service.addExternalSystem(
                sessionId, 
                adapterType, 
                name, 
                title,
                addExtSystemSoapDO)
                
            self.log.info("Registered %s with ExsyId %s" % (name, exsysid))
        except Exception, ex:
            if (self.cfg['SSL'] == 'on'):
                self.logMesg("warn", "\nRegistering SCM integration failed. If you are using self-signed certificates,\n"+
                                     "then please create the SVN integration through TeamForge Admin UI")
            raise Exception(ex)
            
    def __isEndpointExternallyReachable(self, url):
        try:
            urllib2.urlopen(url, timeout=2).getcode()
            return True
            
        except urllib2.HTTPError, e:
            # Any code below 500 means that the server is reachable. A 5xx (503, in particular) could
            # come from a load balancer.
            return e.code < 500
            
        except Exception, e:
            self.log.warn("URL %s not externally reachable (yet): %r" % (url, e))
            return False
            
    def __waitForEndpointToBecomeExternallyReachable(self, url):
        # We know that the integration server is up - but if it is running behind a load balancer,
        # it may not be reachable by the application yet (the load balancer may take a few more seconds
        # before it takes the backend into the pool). However, if the integration server is not
        # reachable by the app, validation will fail and the registration will be rejected.
        #
        # Therefore, give the load balancer some slack.
        for i in range(1, 40):
            if self.__isEndpointExternallyReachable(url):
                return
            else:
                self.log.info("Endpoint '%s' not reachable yet, waiting..." % url)
                time.sleep(1)
                
        raise Exception("URL %s not externally reachable" % url)
        
            
    def __registerSvn(self):
        # Determine own endpoints from the view of the 'ctfcore-app' service (wherever it is deployed)
        coreMachine    = app.landscape.Landscape.getSite().getMachinesByService("ctfcore-app")[0]
        coreAppService = coreMachine.getService("ctfcore-app")
        
        scheme         = 'https' if self.cfg['SSL'] == 'on' else 'http'
        
        soapEndpoint   = coreAppService.getRequiredEndpoint("http-soap", "integration").getProviderEndpoint()
        soapServer     = soapEndpoint.getHost(coreMachine.isLocal())
        soapPort       = soapEndpoint.getPort()

        viewvcEndpoint = coreAppService.getRequiredEndpoint("http-web-viewvc", "integration").getProviderEndpoint()
        viewvcServer   = viewvcEndpoint.getHost(coreMachine.isLocal())
        viewvcPort     = viewvcEndpoint.getPort()
        viewvcUrl      = "%s://%s%s/integration/viewvc/viewvc.cgi" % (scheme, viewvcServer, self.__portSegment(scheme, viewvcPort))
        
        svnEndpoint    = coreAppService.getRequiredEndpoint("http-svn", "integration").getProviderEndpoint()
        svnServer      = svnEndpoint.getHost(coreMachine.isLocal())
        svnPort        = svnEndpoint.getPort()
        svnBaseUrl     = "%s://%s%s%s" % (scheme, svnServer, self.__portSegment(scheme, svnPort), self.cfg['SUBVERSION_URI'])
 
        if soapServer != "localhost":
            self.__waitForEndpointToBecomeExternallyReachable("%s://%s:%d/" % (scheme, soapServer, soapPort))
        
        self.__registerScmIntegration(
            "%s://%s:%s/ce-soap60/services" % (
                "http" if self.cfg["integration/dependency/ctfcore-app/http-soap/host"] == "localhost" else scheme, 
                self.cfg["integration/dependency/ctfcore-app/http-soap/host"], 
                self.cfg["integration/dependency/ctfcore-app/http-soap/port"]),
            'Subversion', 
            "SVN", 
            "TeamForge Subversion SCM integration", 
            ['ScmViewerUrl', 
             'RequireApproval', 
             'RepositoryRootPath',
             'RepositoryBaseUrl', 
             'HostSSL', 
             'HostPort', 
             'HostName',
             'UseInternalCodeBrowser'],
            [viewvcUrl, 
             'false', 
             self.cfg['SUBVERSION_REPOSITORY_BASE'],
             svnBaseUrl, 
             "true" if self.cfg['SSL'] == 'on' and soapServer != "localhost" else "false", 
             soapPort, 
             soapServer,
             'true' ])
            
    def __registerCvs(self):
        # Determine own endpoints from the view of the 'ctfcore-app' service (wherever it is deployed)
        coreMachine    = app.landscape.Landscape.getSite().getMachinesByService("ctfcore-app")[0]
        coreAppService = coreMachine.getService("ctfcore-app")
        
        scheme         = 'https' if self.cfg['SSL'] == 'on' else 'http'
        soapEndpoint   = coreAppService.getRequiredEndpoint("http-soap", "integration").getProviderEndpoint()
        soapServer     = soapEndpoint.getHost(coreMachine.isLocal())
        soapPort       = soapEndpoint.getPort()
        
        viewvcEndpoint = coreAppService.getRequiredEndpoint("http-web-viewvc", "integration").getProviderEndpoint()
        viewvcServer   = viewvcEndpoint.getHost(coreMachine.isLocal())
        viewvcPort     = viewvcEndpoint.getPort()
        viewvcUrl      = "%s://%s%s/integration/viewvc/viewvc.cgi" % (scheme, viewvcServer,  self.__portSegment(scheme, viewvcPort))
    
        if soapServer != "localhost":
            self.__waitForEndpointToBecomeExternallyReachable("%s://%s:%d/" % (scheme, soapServer, soapPort))
        
        self.__registerScmIntegration(  
            "%s://%s:%s/ce-soap60/services" % (
                "http" if self.cfg["integration/dependency/ctfcore-app/http-soap/host"] == "localhost" else scheme, 
                self.cfg["integration/dependency/ctfcore-app/http-soap/host"], 
                self.cfg["integration/dependency/ctfcore-app/http-soap/port"]),
            "CVS", 
            "CVS", 
            "TeamForge CVS SCM integration",  
            ['ScmViewerUrl', 
             'RequireApproval', 
             'RepositoryRootPath',
             'CvsHost', 
             'HostSSL', 
             'HostPort', 
             'HostName' ],
            [viewvcUrl, 
             'false', 
             self.cfg['CVS_REPOSITORY_BASE'],
             soapServer, 
             "true" if self.cfg['SSL'] == 'on' and soapServer != "localhost" else "false", 
             soapPort, 
             soapServer])
    
    def initialize(self):
        if self.cfg.has_key('SUBVERSION_REPOSITORY_BASE') and os.path.isdir(self.cfg['SUBVERSION_REPOSITORY_BASE']) and not self.__isRegistered(self.cfg['SUBVERSION_REPOSITORY_BASE']):
            self.__registerSvn()
            
        if self.cfg.has_key('CVS_REPOSITORY_BASE') and os.path.isdir(self.cfg['CVS_REPOSITORY_BASE']) and not self.__isRegistered(self.cfg['CVS_REPOSITORY_BASE']):
            self.__registerCvs()

